from enum import Enum


class StationarisationStrategy(str, Enum):
    KEEP_FAIL = "keep_fail"
    DISCARD_FAIL = "discard_fail"
    NONE = "none"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.backtest_stage_config import BacktestStageConfig
from ..models.classification_backtest_config import ClassificationBacktestConfig
from ..models.classification_discovery_config import ClassificationDiscoveryConfig
from ..models.classification_specification_config import ClassificationSpecificationConfig
from ..models.feature_generation_stage_config import FeatureGenerationStageConfig
from ..models.filter_stage_config import FilterStageConfig
from ..models.lstm_backtest_stage_config import LstmBacktestStageConfig
from ..models.lstm_prediction_stage_config import LstmPredictionStageConfig
from ..models.prediction_stage_config import PredictionStageConfig
from ..models.problem_specification_config import ProblemSpecificationConfig
from ..models.refinement_stage_config import RefinementStageConfig
from ..models.stationarisation_stage_config import StationarisationStageConfig
from ..models.trading_simulation_stage_config import TradingSimulationStageConfig
from ..models.trading_specification_stage_config import TradingSpecificationStageConfig

T = TypeVar("T", bound="StageConfigUpdate")


@attr.s(auto_attribs=True)
class StageConfigUpdate:
    """  """

    config: Union[
        BacktestStageConfig,
        ClassificationBacktestConfig,
        ClassificationDiscoveryConfig,
        ClassificationSpecificationConfig,
        FeatureGenerationStageConfig,
        FilterStageConfig,
        LstmBacktestStageConfig,
        LstmPredictionStageConfig,
        PredictionStageConfig,
        ProblemSpecificationConfig,
        RefinementStageConfig,
        StationarisationStageConfig,
        TradingSimulationStageConfig,
        TradingSpecificationStageConfig,
    ]
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        if isinstance(self.config, ProblemSpecificationConfig):
            config = self.config.to_dict()

        elif isinstance(self.config, StationarisationStageConfig):
            config = self.config.to_dict()

        elif isinstance(self.config, FeatureGenerationStageConfig):
            config = self.config.to_dict()

        elif isinstance(self.config, FilterStageConfig):
            config = self.config.to_dict()

        elif isinstance(self.config, RefinementStageConfig):
            config = self.config.to_dict()

        elif isinstance(self.config, BacktestStageConfig):
            config = self.config.to_dict()

        elif isinstance(self.config, PredictionStageConfig):
            config = self.config.to_dict()

        elif isinstance(self.config, TradingSpecificationStageConfig):
            config = self.config.to_dict()

        elif isinstance(self.config, TradingSimulationStageConfig):
            config = self.config.to_dict()

        elif isinstance(self.config, ClassificationSpecificationConfig):
            config = self.config.to_dict()

        elif isinstance(self.config, ClassificationDiscoveryConfig):
            config = self.config.to_dict()

        elif isinstance(self.config, ClassificationBacktestConfig):
            config = self.config.to_dict()

        elif isinstance(self.config, LstmBacktestStageConfig):
            config = self.config.to_dict()

        else:
            config = self.config.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "config": config,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()

        def _parse_config(
            data: object,
        ) -> Union[
            BacktestStageConfig,
            ClassificationBacktestConfig,
            ClassificationDiscoveryConfig,
            ClassificationSpecificationConfig,
            FeatureGenerationStageConfig,
            FilterStageConfig,
            LstmBacktestStageConfig,
            LstmPredictionStageConfig,
            PredictionStageConfig,
            ProblemSpecificationConfig,
            RefinementStageConfig,
            StationarisationStageConfig,
            TradingSimulationStageConfig,
            TradingSpecificationStageConfig,
        ]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                config_type_0 = ProblemSpecificationConfig.from_dict(data)

                return config_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                config_type_1 = StationarisationStageConfig.from_dict(data)

                return config_type_1
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                config_type_2 = FeatureGenerationStageConfig.from_dict(data)

                return config_type_2
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                config_type_3 = FilterStageConfig.from_dict(data)

                return config_type_3
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                config_type_4 = RefinementStageConfig.from_dict(data)

                return config_type_4
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                config_type_5 = BacktestStageConfig.from_dict(data)

                return config_type_5
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                config_type_6 = PredictionStageConfig.from_dict(data)

                return config_type_6
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                config_type_7 = TradingSpecificationStageConfig.from_dict(data)

                return config_type_7
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                config_type_8 = TradingSimulationStageConfig.from_dict(data)

                return config_type_8
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                config_type_9 = ClassificationSpecificationConfig.from_dict(data)

                return config_type_9
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                config_type_10 = ClassificationDiscoveryConfig.from_dict(data)

                return config_type_10
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                config_type_11 = ClassificationBacktestConfig.from_dict(data)

                return config_type_11
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                config_type_12 = LstmBacktestStageConfig.from_dict(data)

                return config_type_12
            except:  # noqa: E722
                pass
            if not isinstance(data, dict):
                raise TypeError()
            config_type_13 = LstmPredictionStageConfig.from_dict(data)

            return config_type_13

        config = _parse_config(d.pop("config"))

        stage_config_update = cls(
            config=config,
        )

        stage_config_update.additional_properties = d
        return stage_config_update

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class BlueprintType(str, Enum):
    NONLINEAR = "nonlinear"
    LINEAR = "linear"
    SMALL_DATA = "small_data"
    FAST_FORECASTING = "fast_forecasting"
    LSTM = "lstm"
    TRADING_SIMULATION = "trading_simulation"
    FEATURE_SELECTION = "feature_selection"
    FEATURE_DISCOVERY = "feature_discovery"
    SIGNAL_ENCODING = "signal_encoding"
    VARIATIONAL_FORECASTING = "variational_forecasting"
    STATIONARISATION = "stationarisation"
    TIME_SERIES_REGRESSION = "time_series_regression"
    CLASSIFICATION = "classification"
    CUSTOM = "custom"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class StageStatus(str, Enum):
    READY = "READY"
    RUNNING = "RUNNING"
    COMPLETE = "COMPLETE"
    PENDING = "PENDING"
    ERROR = "ERROR"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.regressor_type import RegressorType
from ..types import UNSET, Unset

T = TypeVar("T", bound="BacktestStageConfig")


@attr.s(auto_attribs=True)
class BacktestStageConfig:
    """ Properties for a stage  """

    n_backtests: int
    fold_train_frac: float
    gapping_factor: float
    regressor: RegressorType
    type: Union[Unset, str] = "backtest"
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        n_backtests = self.n_backtests
        fold_train_frac = self.fold_train_frac
        gapping_factor = self.gapping_factor
        regressor = self.regressor.value

        type = self.type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "nBacktests": n_backtests,
                "foldTrainFrac": fold_train_frac,
                "gappingFactor": gapping_factor,
                "regressor": regressor,
            }
        )
        if type is not UNSET:
            field_dict["type"] = type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        n_backtests = d.pop("nBacktests")

        fold_train_frac = d.pop("foldTrainFrac")

        gapping_factor = d.pop("gappingFactor")

        regressor = RegressorType(d.pop("regressor"))

        type = d.pop("type", UNSET)

        backtest_stage_config = cls(
            n_backtests=n_backtests,
            fold_train_frac=fold_train_frac,
            gapping_factor=gapping_factor,
            regressor=regressor,
            type=type,
        )

        backtest_stage_config.additional_properties = d
        return backtest_stage_config

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.classification_discovery_config import ClassificationDiscoveryConfig
from ..models.stage_status import StageStatus
from ..types import UNSET, Unset

T = TypeVar("T", bound="ClassificationDiscoveryStage")


@attr.s(auto_attribs=True)
class ClassificationDiscoveryStage:
    """Stage for discovering data transformations that might help the classifier

    The stage data will contain the discovered transformations, so other stages can use
    them.

    This Stage is like 'Feature Generation' in regression pipelines, but more flexible,
    allowing for dual-parent features as well as single-parent features."""

    status: StageStatus
    id: int
    config: ClassificationDiscoveryConfig
    error_msg: Union[Unset, None, str] = UNSET
    type: Union[Unset, str] = "classification_discovery"
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        status = self.status.value

        id = self.id
        config = self.config.to_dict()

        error_msg = self.error_msg
        type = self.type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "status": status,
                "id": id,
                "config": config,
            }
        )
        if error_msg is not UNSET:
            field_dict["errorMsg"] = error_msg
        if type is not UNSET:
            field_dict["type"] = type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        status = StageStatus(d.pop("status"))

        id = d.pop("id")

        config = ClassificationDiscoveryConfig.from_dict(d.pop("config"))

        error_msg = d.pop("errorMsg", UNSET)

        type = d.pop("type", UNSET)

        classification_discovery_stage = cls(
            status=status,
            id=id,
            config=config,
            error_msg=error_msg,
            type=type,
        )

        classification_discovery_stage.additional_properties = d
        return classification_discovery_stage

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union, cast

import attr

from ..models.classification_score_metric import ClassificationScoreMetric
from ..models.feature_importance import FeatureImportance
from ..models.summary_score_metric import SummaryScoreMetric
from ..types import UNSET, Unset

T = TypeVar("T", bound="ClassificationBacktestInsight")


@attr.s(auto_attribs=True)
class ClassificationBacktestInsight:
    """Information generated as a side effect of running a stage, detailing
    intermediate results and analyses used to define which features were output from the
    stage."""

    target_label_to_detect: str
    summary_score_metrics: List[SummaryScoreMetric]
    summary_predicted_true_actual_true: int
    summary_predicted_true_actual_false: int
    summary_predicted_false_actual_true: int
    summary_predicted_false_actual_false: int
    average_feature_importances: List[FeatureImportance]
    rolling_score_metric_type: ClassificationScoreMetric
    dates_timestamp_milliseconds: Union[Unset, List[int]] = UNSET
    rolling_score_metric_values: Union[Unset, List[float]] = UNSET
    prediction_pseudo_probabilities: Union[Unset, List[float]] = UNSET
    predictions: Union[Unset, List[bool]] = UNSET
    ground_truths: Union[Unset, List[bool]] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        target_label_to_detect = self.target_label_to_detect
        summary_score_metrics = []
        for summary_score_metrics_item_data in self.summary_score_metrics:
            summary_score_metrics_item = summary_score_metrics_item_data.to_dict()

            summary_score_metrics.append(summary_score_metrics_item)

        summary_predicted_true_actual_true = self.summary_predicted_true_actual_true
        summary_predicted_true_actual_false = self.summary_predicted_true_actual_false
        summary_predicted_false_actual_true = self.summary_predicted_false_actual_true
        summary_predicted_false_actual_false = self.summary_predicted_false_actual_false
        average_feature_importances = []
        for average_feature_importances_item_data in self.average_feature_importances:
            average_feature_importances_item = average_feature_importances_item_data.to_dict()

            average_feature_importances.append(average_feature_importances_item)

        rolling_score_metric_type = self.rolling_score_metric_type.value

        dates_timestamp_milliseconds: Union[Unset, List[int]] = UNSET
        if not isinstance(self.dates_timestamp_milliseconds, Unset):
            dates_timestamp_milliseconds = self.dates_timestamp_milliseconds

        rolling_score_metric_values: Union[Unset, List[float]] = UNSET
        if not isinstance(self.rolling_score_metric_values, Unset):
            rolling_score_metric_values = self.rolling_score_metric_values

        prediction_pseudo_probabilities: Union[Unset, List[float]] = UNSET
        if not isinstance(self.prediction_pseudo_probabilities, Unset):
            prediction_pseudo_probabilities = self.prediction_pseudo_probabilities

        predictions: Union[Unset, List[bool]] = UNSET
        if not isinstance(self.predictions, Unset):
            predictions = self.predictions

        ground_truths: Union[Unset, List[bool]] = UNSET
        if not isinstance(self.ground_truths, Unset):
            ground_truths = self.ground_truths

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "targetLabelToDetect": target_label_to_detect,
                "summaryScoreMetrics": summary_score_metrics,
                "summaryPredictedTrueActualTrue": summary_predicted_true_actual_true,
                "summaryPredictedTrueActualFalse": summary_predicted_true_actual_false,
                "summaryPredictedFalseActualTrue": summary_predicted_false_actual_true,
                "summaryPredictedFalseActualFalse": summary_predicted_false_actual_false,
                "averageFeatureImportances": average_feature_importances,
                "rollingScoreMetricType": rolling_score_metric_type,
            }
        )
        if dates_timestamp_milliseconds is not UNSET:
            field_dict["datesTimestampMilliseconds"] = dates_timestamp_milliseconds
        if rolling_score_metric_values is not UNSET:
            field_dict["rollingScoreMetricValues"] = rolling_score_metric_values
        if prediction_pseudo_probabilities is not UNSET:
            field_dict["predictionPseudoProbabilities"] = prediction_pseudo_probabilities
        if predictions is not UNSET:
            field_dict["predictions"] = predictions
        if ground_truths is not UNSET:
            field_dict["groundTruths"] = ground_truths

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        target_label_to_detect = d.pop("targetLabelToDetect")

        summary_score_metrics = []
        _summary_score_metrics = d.pop("summaryScoreMetrics")
        for summary_score_metrics_item_data in _summary_score_metrics:
            summary_score_metrics_item = SummaryScoreMetric.from_dict(
                summary_score_metrics_item_data
            )

            summary_score_metrics.append(summary_score_metrics_item)

        summary_predicted_true_actual_true = d.pop("summaryPredictedTrueActualTrue")

        summary_predicted_true_actual_false = d.pop("summaryPredictedTrueActualFalse")

        summary_predicted_false_actual_true = d.pop("summaryPredictedFalseActualTrue")

        summary_predicted_false_actual_false = d.pop("summaryPredictedFalseActualFalse")

        average_feature_importances = []
        _average_feature_importances = d.pop("averageFeatureImportances")
        for average_feature_importances_item_data in _average_feature_importances:
            average_feature_importances_item = FeatureImportance.from_dict(
                average_feature_importances_item_data
            )

            average_feature_importances.append(average_feature_importances_item)

        rolling_score_metric_type = ClassificationScoreMetric(d.pop("rollingScoreMetricType"))

        dates_timestamp_milliseconds = cast(List[int], d.pop("datesTimestampMilliseconds", UNSET))

        rolling_score_metric_values = cast(List[float], d.pop("rollingScoreMetricValues", UNSET))

        prediction_pseudo_probabilities = cast(
            List[float], d.pop("predictionPseudoProbabilities", UNSET)
        )

        predictions = cast(List[bool], d.pop("predictions", UNSET))

        ground_truths = cast(List[bool], d.pop("groundTruths", UNSET))

        classification_backtest_insight = cls(
            target_label_to_detect=target_label_to_detect,
            summary_score_metrics=summary_score_metrics,
            summary_predicted_true_actual_true=summary_predicted_true_actual_true,
            summary_predicted_true_actual_false=summary_predicted_true_actual_false,
            summary_predicted_false_actual_true=summary_predicted_false_actual_true,
            summary_predicted_false_actual_false=summary_predicted_false_actual_false,
            average_feature_importances=average_feature_importances,
            rolling_score_metric_type=rolling_score_metric_type,
            dates_timestamp_milliseconds=dates_timestamp_milliseconds,
            rolling_score_metric_values=rolling_score_metric_values,
            prediction_pseudo_probabilities=prediction_pseudo_probabilities,
            predictions=predictions,
            ground_truths=ground_truths,
        )

        classification_backtest_insight.additional_properties = d
        return classification_backtest_insight

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="RawColumn")


@attr.s(auto_attribs=True)
class RawColumn:
    """ Represents raw data, as defined in a dataset on upload  """

    id: int
    name: str
    is_text: Union[Unset, bool] = False
    is_binary: Union[Unset, bool] = False
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        name = self.name
        is_text = self.is_text
        is_binary = self.is_binary

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "name": name,
            }
        )
        if is_text is not UNSET:
            field_dict["isText"] = is_text
        if is_binary is not UNSET:
            field_dict["isBinary"] = is_binary

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        name = d.pop("name")

        is_text = d.pop("isText", UNSET)

        is_binary = d.pop("isBinary", UNSET)

        raw_column = cls(
            id=id,
            name=name,
            is_text=is_text,
            is_binary=is_binary,
        )

        raw_column.additional_properties = d
        return raw_column

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import io
import time

import pandas as pd

from mindfoundry.client.horizon.generated import Client
from mindfoundry.client.horizon.generated.api.datasets import (
    delete_individual_dataset_resource,
    get_data_slice_resource,
    get_dataset_resource,
    get_individual_dataset_resource,
    post_dataset_append_resource,
    post_dataset_upload_resource,
)
from mindfoundry.client.horizon.generated.api.ingestion_processes import (
    get_individual_ingestion_process_resource,
)
from mindfoundry.client.horizon.generated.models import (
    AllDatasets,
    FileUploadForm,
    IndividualDataset,
    IngestionProcess,
    IngestionStatus,
)
from mindfoundry.client.horizon.generated.types import File

from .utils import (
    DEFAULT_UPLOAD_OPTIONS,
    assert_success,
    data_frame_to_buffer,
    return_value_or_raise_error,
)


class DatasetsClient:
    def __init__(self, client: Client):
        self._client = client

    def get_all(self) -> AllDatasets:
        datasets = get_dataset_resource.sync_detailed(client=self._client)
        return return_value_or_raise_error(datasets)

    def get(self, dataset_id: int) -> IndividualDataset:
        dataset = get_individual_dataset_resource.sync_detailed(
            client=self._client,
            id=dataset_id,
        )
        return return_value_or_raise_error(dataset)

    def delete(self, dataset_id: int) -> None:
        response = delete_individual_dataset_resource.sync_detailed(
            client=self._client,
            id=dataset_id,
        )
        assert_success(response)

    def get_data(self, dataset_id: int, start: int, stop: int) -> pd.DataFrame:
        all_data = get_data_slice_resource.sync_detailed(
            client=self._client,
            id=dataset_id,
            start=start,
            stop=stop,
        )
        file = return_value_or_raise_error(all_data)
        return pd.read_csv(file.payload, encoding="utf-8")

    def upload(self, df: pd.DataFrame, name: str) -> IndividualDataset:
        str_buffer = data_frame_to_buffer(df)
        ingestion_process = self._upload_csv(str_buffer, name)
        ingestion_process = self._wait_for_ingestion_process(ingestion_process.id)
        return self._resolve_ingestion_process(ingestion_process)

    def append(self, dataset_id: int, df: pd.DataFrame) -> IndividualDataset:
        str_buffer = data_frame_to_buffer(df)
        ingestion_process = self._append_csv(str_buffer, dataset_id)
        ingestion_process = self._wait_for_ingestion_process(ingestion_process.id)
        return self._resolve_ingestion_process(ingestion_process)

    def _append_csv(self, str_buffer: io.StringIO, dataset_id: int) -> IngestionProcess:
        ingestion_process = post_dataset_append_resource.sync_detailed(
            client=self._client,
            id=dataset_id,
            multipart_data=FileUploadForm(
                options=DEFAULT_UPLOAD_OPTIONS,
                file=File(
                    payload=str_buffer,
                    file_name=f"{dataset_id}.csv",
                    mime_type="text/csv",
                ),
            ),
        )
        return return_value_or_raise_error(ingestion_process)

    def _upload_csv(self, str_buffer: io.StringIO, name: str) -> IngestionProcess:
        ingestion_process = post_dataset_upload_resource.sync_detailed(
            client=self._client,
            multipart_data=FileUploadForm(
                options=DEFAULT_UPLOAD_OPTIONS,
                file=File(
                    payload=str_buffer,
                    file_name=name,
                    mime_type="text/csv",
                ),
            ),
        )
        return return_value_or_raise_error(ingestion_process)

    def _get_ingestion_process(self, ingestion_process_id: int) -> IngestionProcess:
        ingestion_process = get_individual_ingestion_process_resource.sync_detailed(
            client=self._client,
            id=ingestion_process_id,
        )
        return return_value_or_raise_error(ingestion_process)

    def _wait_for_ingestion_process(self, ingestion_process_id: int) -> IngestionProcess:
        ingestion_process = self._get_ingestion_process(ingestion_process_id)

        while ingestion_process.status not in (IngestionStatus.COMPLETED, IngestionStatus.ERROR):
            time.sleep(2)
            ingestion_process = self._get_ingestion_process(ingestion_process_id)

        return ingestion_process

    def _resolve_ingestion_process(self, ingestion_process: IngestionProcess) -> IndividualDataset:
        if ingestion_process.status == IngestionStatus.ERROR:
            raise RuntimeError(ingestion_process.error)

        if ingestion_process.last_update_error is not None:
            raise RuntimeError(ingestion_process.last_update_error)

        if ingestion_process.status == IngestionStatus.COMPLETED:
            assert isinstance(ingestion_process.dataset_id, int)
            return self.get(ingestion_process.dataset_id)

        raise RuntimeError(f"Invalid ingestion status reached: {ingestion_process.status}")



from typing import Any, Dict, List, Type, TypeVar, Union, cast

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="LabelFrequenciesGroupedByTime")


@attr.s(auto_attribs=True)
class LabelFrequenciesGroupedByTime:
    """  """

    dates_milliseconds: Union[Unset, List[int]] = UNSET
    labels: Union[Unset, List[str]] = UNSET
    frequencies_by_date_by_label: Union[Unset, List[List[int]]] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        dates_milliseconds: Union[Unset, List[int]] = UNSET
        if not isinstance(self.dates_milliseconds, Unset):
            dates_milliseconds = self.dates_milliseconds

        labels: Union[Unset, List[str]] = UNSET
        if not isinstance(self.labels, Unset):
            labels = self.labels

        frequencies_by_date_by_label: Union[Unset, List[List[int]]] = UNSET
        if not isinstance(self.frequencies_by_date_by_label, Unset):
            frequencies_by_date_by_label = []
            for frequencies_by_date_by_label_item_data in self.frequencies_by_date_by_label:
                frequencies_by_date_by_label_item = frequencies_by_date_by_label_item_data

                frequencies_by_date_by_label.append(frequencies_by_date_by_label_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if dates_milliseconds is not UNSET:
            field_dict["datesMilliseconds"] = dates_milliseconds
        if labels is not UNSET:
            field_dict["labels"] = labels
        if frequencies_by_date_by_label is not UNSET:
            field_dict["frequenciesByDateByLabel"] = frequencies_by_date_by_label

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        dates_milliseconds = cast(List[int], d.pop("datesMilliseconds", UNSET))

        labels = cast(List[str], d.pop("labels", UNSET))

        frequencies_by_date_by_label = []
        _frequencies_by_date_by_label = d.pop("frequenciesByDateByLabel", UNSET)
        for frequencies_by_date_by_label_item_data in _frequencies_by_date_by_label or []:
            frequencies_by_date_by_label_item = cast(
                List[int], frequencies_by_date_by_label_item_data
            )

            frequencies_by_date_by_label.append(frequencies_by_date_by_label_item)

        label_frequencies_grouped_by_time = cls(
            dates_milliseconds=dates_milliseconds,
            labels=labels,
            frequencies_by_date_by_label=frequencies_by_date_by_label,
        )

        label_frequencies_grouped_by_time.additional_properties = d
        return label_frequencies_grouped_by_time

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from .base import ModelConfig, ValidatedConfig, assert_not_empty_if_provided
from .stages import (
    ClassificationDiscoveryConfig,
    FeatureGenerationConfig,
    FilteringConfig,
    RefinementConfig,
    StationarizationConfig,
)
from .values import (
    CorrelationMethod,
    FeatureGeneratorType,
    RegressorType,
    StationarizationStrategy,
    TargetTransformType,
)



from typing import Any, Dict, List, Type, TypeVar

import attr

from ..models.column_passport import ColumnPassport
from ..models.dataset import Dataset

T = TypeVar("T", bound="IndividualDataset")


@attr.s(auto_attribs=True)
class IndividualDataset:
    """ Dataset summary info and analysis of its columns  """

    analysis: List[ColumnPassport]
    summary: Dataset
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        analysis = []
        for analysis_item_data in self.analysis:
            analysis_item = analysis_item_data.to_dict()

            analysis.append(analysis_item)

        summary = self.summary.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "analysis": analysis,
                "summary": summary,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        analysis = []
        _analysis = d.pop("analysis")
        for analysis_item_data in _analysis:
            analysis_item = ColumnPassport.from_dict(analysis_item_data)

            analysis.append(analysis_item)

        summary = Dataset.from_dict(d.pop("summary"))

        individual_dataset = cls(
            analysis=analysis,
            summary=summary,
        )

        individual_dataset.additional_properties = d
        return individual_dataset

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar

import attr

from ..models.pandas_series import PandasSeries
from ..models.predictions_dto_regressorimportances import PredictionsDTORegressorimportances

T = TypeVar("T", bound="PredictionsDTO")


@attr.s(auto_attribs=True)
class PredictionsDTO:
    """  """

    mean: PandasSeries
    cb_low: PandasSeries
    cb_high: PandasSeries
    confidence: float
    regressor_importances: PredictionsDTORegressorimportances
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        mean = self.mean.to_dict()

        cb_low = self.cb_low.to_dict()

        cb_high = self.cb_high.to_dict()

        confidence = self.confidence
        regressor_importances = self.regressor_importances.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "mean": mean,
                "cbLow": cb_low,
                "cbHigh": cb_high,
                "confidence": confidence,
                "regressorImportances": regressor_importances,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        mean = PandasSeries.from_dict(d.pop("mean"))

        cb_low = PandasSeries.from_dict(d.pop("cbLow"))

        cb_high = PandasSeries.from_dict(d.pop("cbHigh"))

        confidence = d.pop("confidence")

        regressor_importances = PredictionsDTORegressorimportances.from_dict(
            d.pop("regressorImportances")
        )

        predictions_dto = cls(
            mean=mean,
            cb_low=cb_low,
            cb_high=cb_high,
            confidence=confidence,
            regressor_importances=regressor_importances,
        )

        predictions_dto.additional_properties = d
        return predictions_dto

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class IngestionStatus(str, Enum):
    PENDING = "pending"
    IN_PROGRESS = "in_progress"
    COMPLETED = "completed"
    ERROR = "error"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, List, Type, TypeVar, cast

import attr

from ..models.feature_transform import FeatureTransform

T = TypeVar("T", bound="NodeAndLink")


@attr.s(auto_attribs=True)
class NodeAndLink:
    """A graph representation of features. Each feature is a node.
    Each edge is a transformation to compute a feature from other features"""

    feature_id: str
    name: str
    horizon: int
    active: bool
    original_column_names: List[str]
    original_column_ids: List[str]
    immediate_parent_ids: List[str]
    transform: FeatureTransform
    pearson_correlation_with_target: float
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        feature_id = self.feature_id
        name = self.name
        horizon = self.horizon
        active = self.active
        original_column_names = self.original_column_names

        original_column_ids = self.original_column_ids

        immediate_parent_ids = self.immediate_parent_ids

        transform = self.transform.to_dict()

        pearson_correlation_with_target = self.pearson_correlation_with_target

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "featureId": feature_id,
                "name": name,
                "horizon": horizon,
                "active": active,
                "originalColumnNames": original_column_names,
                "originalColumnIds": original_column_ids,
                "immediateParentIds": immediate_parent_ids,
                "transform": transform,
                "pearsonCorrelationWithTarget": pearson_correlation_with_target,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        feature_id = d.pop("featureId")

        name = d.pop("name")

        horizon = d.pop("horizon")

        active = d.pop("active")

        original_column_names = cast(List[str], d.pop("originalColumnNames"))

        original_column_ids = cast(List[str], d.pop("originalColumnIds"))

        immediate_parent_ids = cast(List[str], d.pop("immediateParentIds"))

        transform = FeatureTransform.from_dict(d.pop("transform"))

        pearson_correlation_with_target = d.pop("pearsonCorrelationWithTarget")

        node_and_link = cls(
            feature_id=feature_id,
            name=name,
            horizon=horizon,
            active=active,
            original_column_names=original_column_names,
            original_column_ids=original_column_ids,
            immediate_parent_ids=immediate_parent_ids,
            transform=transform,
            pearson_correlation_with_target=pearson_correlation_with_target,
        )

        node_and_link.additional_properties = d
        return node_and_link

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Optional

import httpx

from ...client import Client
from ...models.expert_predict_form import ExpertPredictForm
from ...models.predict_response_list import PredictResponseList
from ...types import Response


def _get_kwargs(
    *,
    client: Client,
    id: int,
    multipart_data: ExpertPredictForm,
) -> Dict[str, Any]:
    url = "{}/pipelines/{id}/expert/predict".format(client.base_url, id=id)

    headers: Dict[str, Any] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    multipart_multipart_data = multipart_data.to_multipart()

    return {
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
        "files": multipart_multipart_data,
    }


def _parse_response(*, response: httpx.Response) -> Optional[PredictResponseList]:
    if response.status_code == 200:
        response_200 = PredictResponseList.from_dict(response.json())

        return response_200
    return None


def _build_response(*, response: httpx.Response) -> Response[PredictResponseList]:
    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(response=response),
    )


def sync_detailed(
    *,
    client: Client,
    id: int,
    multipart_data: ExpertPredictForm,
) -> Response[PredictResponseList]:
    kwargs = _get_kwargs(
        client=client,
        id=id,
        multipart_data=multipart_data,
    )

    response = httpx.post(
        **kwargs,
    )

    return _build_response(response=response)


def sync(
    *,
    client: Client,
    id: int,
    multipart_data: ExpertPredictForm,
) -> Optional[PredictResponseList]:
    """  """

    return sync_detailed(
        client=client,
        id=id,
        multipart_data=multipart_data,
    ).parsed


async def asyncio_detailed(
    *,
    client: Client,
    id: int,
    multipart_data: ExpertPredictForm,
) -> Response[PredictResponseList]:
    kwargs = _get_kwargs(
        client=client,
        id=id,
        multipart_data=multipart_data,
    )

    async with httpx.AsyncClient() as _client:
        response = await _client.post(**kwargs)

    return _build_response(response=response)


async def asyncio(
    *,
    client: Client,
    id: int,
    multipart_data: ExpertPredictForm,
) -> Optional[PredictResponseList]:
    """  """

    return (
        await asyncio_detailed(
            client=client,
            id=id,
            multipart_data=multipart_data,
        )
    ).parsed



import abc
import uuid
from dataclasses import dataclass
from typing import Any, List, Optional, Union

from mindfoundry.client.horizon.generated.models import IndividualDataset


class ValidatedConfig(abc.ABC):
    @abc.abstractmethod
    def validate(self) -> None:
        pass


@dataclass
class ModelConfig(ValidatedConfig):
    name: Optional[str] = None
    """
    Name of the model. Although not necessary, it is advised to give models a name so that
    multiple models can be easily told apart. If not provided, the model will be given a
    random name.
    """

    inputs: Optional[List[str]] = None
    """
    List of column names to be used as features for the forecasting. Only the specified columns
    will be processed and analyzed. Any column that is not in this list will be ignored.
    By default, all columns will be used.
    Mutually exclusive with `exclude_from_input` (use one or the other but not both).
    """

    exclude_from_input: Optional[List[str]] = None
    """
    List of column names to be ignored in the forecasting. Every other column that is present in
    the data but not specified in this list will be used as a feature.
    By default, no column will be excluded.
    Mutually exclusive with `inputs` (use one or the other but not both).
    """

    def validate(self) -> None:
        assert self.inputs is None or self.exclude_from_input is None, (
            "inputs and exclude_from_input cannot be specified together at the same time."
            "Use one or the other"
        )

        assert_not_empty_if_provided("name", self.name)
        assert_not_empty_if_provided("inputs", self.inputs, True)
        assert_not_empty_if_provided("exclude_from_input", self.exclude_from_input, True)

    def resolve_dataset_name(self) -> str:
        code = str(uuid.uuid4())
        return self.name or f"Dataset {code[:8]}"

    def resolve_pipeline_name(self) -> str:
        code = str(uuid.uuid4())
        return self.name or f"Pipeline {code[:8]}"

    def resolve_active_columns(self, dataset: IndividualDataset) -> List[int]:
        if self.inputs is not None:
            assert_no_missing_columns(dataset, self.inputs)
            return [cp.id for cp in dataset.analysis if cp.name in self.inputs]

        if self.exclude_from_input is not None:
            assert_no_missing_columns(dataset, self.exclude_from_input)
            return [cp.id for cp in dataset.analysis if cp.name not in self.exclude_from_input]

        return [cp.id for cp in dataset.analysis]


def assert_not_empty_if_provided(
    field: str,
    value: Optional[Union[str, List[Any]]] = None,
    must_be_list: Optional[bool] = False,
) -> None:
    if value is not None:
        assert len(value) > 0, f"{field} cannot be empty"
        if must_be_list:
            assert isinstance(value, List), f"{field} must be a List"


def assert_no_missing_columns(dataset: IndividualDataset, wanted_names: List[str]) -> None:
    existing_column_names = [cp.name for cp in dataset.analysis]
    missing_columns = [name for name in wanted_names if name not in existing_column_names]
    assert len(missing_columns) == 0, (
        "Some column names were used that are not actually present in the dataset. "
        f"Please review the following: {missing_columns}"
    )



from typing import Any, Dict, List, Type, TypeVar, cast

import attr

from ..models.feature_transform_params import FeatureTransformParams

T = TypeVar("T", bound="FeatureTransform")


@attr.s(auto_attribs=True)
class FeatureTransform:
    """  """

    secondary_parameters: List[str]
    transform_category: str
    params: FeatureTransformParams
    explanation: str
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        secondary_parameters = self.secondary_parameters

        transform_category = self.transform_category
        params = self.params.to_dict()

        explanation = self.explanation

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "secondaryParameters": secondary_parameters,
                "transformCategory": transform_category,
                "params": params,
                "explanation": explanation,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        secondary_parameters = cast(List[str], d.pop("secondaryParameters"))

        transform_category = d.pop("transformCategory")

        params = FeatureTransformParams.from_dict(d.pop("params"))

        explanation = d.pop("explanation")

        feature_transform = cls(
            secondary_parameters=secondary_parameters,
            transform_category=transform_category,
            params=params,
            explanation=explanation,
        )

        feature_transform.additional_properties = d
        return feature_transform

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, cast

import attr

from ..models.json_data_dtype import JsonDataDtype

T = TypeVar("T", bound="JsonData")


@attr.s(auto_attribs=True)
class JsonData:
    """  """

    name: str
    columns: List[JsonDataDtype]
    rows: List[List[str]]
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        columns = []
        for columns_item_data in self.columns:
            columns_item = columns_item_data.to_dict()

            columns.append(columns_item)

        rows = []
        for rows_item_data in self.rows:
            rows_item = rows_item_data

            rows.append(rows_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "columns": columns,
                "rows": rows,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        columns = []
        _columns = d.pop("columns")
        for columns_item_data in _columns:
            columns_item = JsonDataDtype.from_dict(columns_item_data)

            columns.append(columns_item)

        rows = []
        _rows = d.pop("rows")
        for rows_item_data in _rows:
            rows_item = cast(List[str], rows_item_data)

            rows.append(rows_item)

        json_data = cls(
            name=name,
            columns=columns,
            rows=rows,
        )

        json_data.additional_properties = d
        return json_data

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union, cast

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="ClassificationDiscoveryConfig")


@attr.s(auto_attribs=True)
class ClassificationDiscoveryConfig:
    """ Properties for a stage  """

    feature_generation_enabled: bool
    timeout_seconds: float
    columns_to_not_generate_features_from: List[str]
    max_n_features: int
    type: Union[Unset, str] = "classification_discovery"
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        feature_generation_enabled = self.feature_generation_enabled
        timeout_seconds = self.timeout_seconds
        columns_to_not_generate_features_from = self.columns_to_not_generate_features_from

        max_n_features = self.max_n_features
        type = self.type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "featureGenerationEnabled": feature_generation_enabled,
                "timeoutSeconds": timeout_seconds,
                "columnsToNotGenerateFeaturesFrom": columns_to_not_generate_features_from,
                "maxNFeatures": max_n_features,
            }
        )
        if type is not UNSET:
            field_dict["type"] = type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        feature_generation_enabled = d.pop("featureGenerationEnabled")

        timeout_seconds = d.pop("timeoutSeconds")

        columns_to_not_generate_features_from = cast(
            List[str], d.pop("columnsToNotGenerateFeaturesFrom")
        )

        max_n_features = d.pop("maxNFeatures")

        type = d.pop("type", UNSET)

        classification_discovery_config = cls(
            feature_generation_enabled=feature_generation_enabled,
            timeout_seconds=timeout_seconds,
            columns_to_not_generate_features_from=columns_to_not_generate_features_from,
            max_n_features=max_n_features,
            type=type,
        )

        classification_discovery_config.additional_properties = d
        return classification_discovery_config

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class FeatureGeneratorType(Enum):
    """
    Contains a variety of techniques used to generate new features from a dataset's columns.
    """

    AUTO_LAG = "autolag"
    """ Creates optimal lag transforms based on the auto-correlation of a feature. """

    EWMA = "ewma"
    """ Exponentially weighted moving average. """

    LAG = "lag"
    """
    Produces a delayed copy of the same feature, where the delay is based on the series
    `total` length.
    """

    LOGARITHM = "logarithm"
    """ Applies the natural logarithm function to all values in a feature. """

    ROLLING_AVERAGE = "rolling_average"
    """ Applies a rolling average transform on a feature. """

    CALENDAR = "calendar"
    """ Extracts a new feature based on the calendar components of the time index. """

    PERCENT_CHANGE = "perc_change"
    """ Produces a percent change of a feature along specific lags. """

    NUMBER_OF_PEAKS = "num_peaks"
    """ Counts the number of peaks in recent data. """

    ONE_HOT_ENCODING = "one_hot_encode"
    """ Used to convert non-numeric data into numeric data. """


class StationarizationStrategy(Enum):
    """ Describe how to treat results of the stationarization stage. """

    KEEP_FAILED = "keep_fail"
    """ Keep original columns even if they cannot be stationarized. """

    DISCARD_FAILED = "discard_fail"
    """ Drop columns that cannot be stationarized. """

    NONE = "none"
    """ Do nothing, return original columns. """


class TargetTransformType(Enum):
    """ Indicates whether target columns should be transformed. """

    DO_NOTHING = "DoNothing"
    """ Columns are kept as they originally appear in the data. """

    HORIZON_LAG_DIFF = "HorizonLagDiff"
    """ Applies a difference function across a lag equal to the current horizon. """

    HORIZON_LAG_DIFF_RATIO = "HorizonLagDiffRatio"
    """ Similar to the above, but taken as a ratio between values. """


class CorrelationMethod(Enum):
    """ Describes which metric is used to compute correlation between features. """

    MUTUAL_INFORMATION = "mutual_info"
    SPEARMAN = "spearman"
    PEARSON = "pearson"
    KENDALL = "kendall"


class RegressorType(Enum):
    """
    Indicates which regression algorithm is used to produce a regression model.
    """

    RANDOM_FOREST = "RandomForest"
    MARTINGALE = "Martingale"
    VBLINREG = "VBLinReg"
    MONDRIAN_FOREST = "MondrianForest"
    XGBOOST = "XGBoost"



from typing import Any, Dict, List, Type, TypeVar, Union, cast

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="PandasSeries")


@attr.s(auto_attribs=True)
class PandasSeries:
    """  """

    data: List[Union[float, str]]
    index: List[float]
    index_dtype: str
    name: Union[Unset, None, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        data = []
        for data_item_data in self.data:
            data_item = data_item_data

            data.append(data_item)

        index = []
        for index_item_data in self.index:
            index_item = index_item_data

            index.append(index_item)

        index_dtype = self.index_dtype
        name = self.name

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "data": data,
                "index": index,
                "indexDtype": index_dtype,
            }
        )
        if name is not UNSET:
            field_dict["name"] = name

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        data = []
        _data = d.pop("data")
        for data_item_data in _data:

            def _parse_data_item(data: object) -> Union[float, str]:
                return cast(Union[float, str], data)

            data_item = _parse_data_item(data_item_data)

            data.append(data_item)

        index = []
        _index = d.pop("index")
        for index_item_data in _index:

            def _parse_index_item(data: object) -> float:
                return cast(float, data)

            index_item = _parse_index_item(index_item_data)

            index.append(index_item)

        index_dtype = d.pop("indexDtype")

        name = d.pop("name", UNSET)

        pandas_series = cls(
            data=data,
            index=index,
            index_dtype=index_dtype,
            name=name,
        )

        pandas_series.additional_properties = d
        return pandas_series

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.refinement_stage_config import RefinementStageConfig
from ..models.stage_status import StageStatus
from ..types import UNSET, Unset

T = TypeVar("T", bound="RefinementStage")


@attr.s(auto_attribs=True)
class RefinementStage:
    """ Further reduction of feature set, based on how each subset performs in a regression model """

    status: StageStatus
    id: int
    config: RefinementStageConfig
    error_msg: Union[Unset, None, str] = UNSET
    type: Union[Unset, str] = "refinement"
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        status = self.status.value

        id = self.id
        config = self.config.to_dict()

        error_msg = self.error_msg
        type = self.type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "status": status,
                "id": id,
                "config": config,
            }
        )
        if error_msg is not UNSET:
            field_dict["errorMsg"] = error_msg
        if type is not UNSET:
            field_dict["type"] = type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        status = StageStatus(d.pop("status"))

        id = d.pop("id")

        config = RefinementStageConfig.from_dict(d.pop("config"))

        error_msg = d.pop("errorMsg", UNSET)

        type = d.pop("type", UNSET)

        refinement_stage = cls(
            status=status,
            id=id,
            config=config,
            error_msg=error_msg,
            type=type,
        )

        refinement_stage.additional_properties = d
        return refinement_stage

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar

import attr

T = TypeVar("T", bound="OHLCPointSchema")


@attr.s(auto_attribs=True)
class OHLCPointSchema:
    """  """

    open_: float
    close: float
    high: float
    low: float
    date: str
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        open_ = self.open_
        close = self.close
        high = self.high
        low = self.low
        date = self.date

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "open": open_,
                "close": close,
                "high": high,
                "low": low,
                "date": date,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        open_ = d.pop("open")

        close = d.pop("close")

        high = d.pop("high")

        low = d.pop("low")

        date = d.pop("date")

        ohlc_point_schema = cls(
            open_=open_,
            close=close,
            high=high,
            low=low,
            date=date,
        )

        ohlc_point_schema.additional_properties = d
        return ohlc_point_schema

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class RegressorType(str, Enum):
    RANDOMFOREST = "RandomForest"
    MARTINGALE = "Martingale"
    VBLINREG = "VBLinReg"
    MONDRIANFOREST = "MondrianForest"
    XGBOOST = "XGBoost"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, Optional

import httpx

from ...client import Client
from ...models.file_upload_form import FileUploadForm
from ...models.ingestion_process import IngestionProcess
from ...types import Response


def _get_kwargs(
    *,
    client: Client,
    id: int,
    multipart_data: FileUploadForm,
) -> Dict[str, Any]:
    url = "{}/datasets/{id}/append".format(client.base_url, id=id)

    headers: Dict[str, Any] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    multipart_multipart_data = multipart_data.to_multipart()

    return {
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
        "files": multipart_multipart_data,
    }


def _parse_response(*, response: httpx.Response) -> Optional[IngestionProcess]:
    if response.status_code == 200:
        response_200 = IngestionProcess.from_dict(response.json())

        return response_200
    return None


def _build_response(*, response: httpx.Response) -> Response[IngestionProcess]:
    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(response=response),
    )


def sync_detailed(
    *,
    client: Client,
    id: int,
    multipart_data: FileUploadForm,
) -> Response[IngestionProcess]:
    kwargs = _get_kwargs(
        client=client,
        id=id,
        multipart_data=multipart_data,
    )

    response = httpx.post(
        **kwargs,
    )

    return _build_response(response=response)


def sync(
    *,
    client: Client,
    id: int,
    multipart_data: FileUploadForm,
) -> Optional[IngestionProcess]:
    """Request body, upload options and return object are identical to the upload endpoint's.
    All of the following conditions must be met in order for the process to succeed:

    - The current user did not exhaust their data allowance;
    - The current user must be the same user that created the dataset in the first place
      (or a support user);
    - There are no running pipelines that use the given dataset `dataset_id`;
    - The upload process of `dataset_id` must be completed, i.e. there cannot be any
      non-completed ingestion process for that dataset;
    - New data must contain exactly the same columns of the same types w.r.t. existing
      data currently stored in the dataset;
    - ALL new data points must be at least as recent as the most recent row in the dataset;
    - If the dataset has unique indices, then new data must also have unique indices;
    - New data must have the same cadence as existing data;
    - The oldest data points in the new data must be exactly 0 or 1 cadence away from
      the most recent row in the dataset.

    If any one of the conditions above is not met, this method will fail with an error.
    The corresponding ingestion process will be still marked as completed, but its
    `last_update_error` field will report the same error. The dataset will still be accessible.
    """

    return sync_detailed(
        client=client,
        id=id,
        multipart_data=multipart_data,
    ).parsed


async def asyncio_detailed(
    *,
    client: Client,
    id: int,
    multipart_data: FileUploadForm,
) -> Response[IngestionProcess]:
    kwargs = _get_kwargs(
        client=client,
        id=id,
        multipart_data=multipart_data,
    )

    async with httpx.AsyncClient() as _client:
        response = await _client.post(**kwargs)

    return _build_response(response=response)


async def asyncio(
    *,
    client: Client,
    id: int,
    multipart_data: FileUploadForm,
) -> Optional[IngestionProcess]:
    """Request body, upload options and return object are identical to the upload endpoint's.
    All of the following conditions must be met in order for the process to succeed:

    - The current user did not exhaust their data allowance;
    - The current user must be the same user that created the dataset in the first place
      (or a support user);
    - There are no running pipelines that use the given dataset `dataset_id`;
    - The upload process of `dataset_id` must be completed, i.e. there cannot be any
      non-completed ingestion process for that dataset;
    - New data must contain exactly the same columns of the same types w.r.t. existing
      data currently stored in the dataset;
    - ALL new data points must be at least as recent as the most recent row in the dataset;
    - If the dataset has unique indices, then new data must also have unique indices;
    - New data must have the same cadence as existing data;
    - The oldest data points in the new data must be exactly 0 or 1 cadence away from
      the most recent row in the dataset.

    If any one of the conditions above is not met, this method will fail with an error.
    The corresponding ingestion process will be still marked as completed, but its
    `last_update_error` field will report the same error. The dataset will still be accessible.
    """

    return (
        await asyncio_detailed(
            client=client,
            id=id,
            multipart_data=multipart_data,
        )
    ).parsed



from typing import Any, Dict, List, Type, TypeVar

import attr

from ..models.algo_execution_test_case import AlgoExecutionTestCase

T = TypeVar("T", bound="AlgoExecutionTestCaseList")


@attr.s(auto_attribs=True)
class AlgoExecutionTestCaseList:
    """  """

    test_cases: List[AlgoExecutionTestCase]
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        test_cases = []
        for test_cases_item_data in self.test_cases:
            test_cases_item = test_cases_item_data.to_dict()

            test_cases.append(test_cases_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "testCases": test_cases,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        test_cases = []
        _test_cases = d.pop("testCases")
        for test_cases_item_data in _test_cases:
            test_cases_item = AlgoExecutionTestCase.from_dict(test_cases_item_data)

            test_cases.append(test_cases_item)

        algo_execution_test_case_list = cls(
            test_cases=test_cases,
        )

        algo_execution_test_case_list.additional_properties = d
        return algo_execution_test_case_list

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar

import attr

from ..models.serialiser_name import SerialiserName
from ..models.storage_engine import StorageEngine

T = TypeVar("T", bound="StorageSpecification")


@attr.s(auto_attribs=True)
class StorageSpecification:
    """ Details on how the data is stored """

    engine: StorageEngine
    chunk_size: int
    serialiser: SerialiserName
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        engine = self.engine.value

        chunk_size = self.chunk_size
        serialiser = self.serialiser.value

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "engine": engine,
                "chunkSize": chunk_size,
                "serialiser": serialiser,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        engine = StorageEngine(d.pop("engine"))

        chunk_size = d.pop("chunkSize")

        serialiser = SerialiserName(d.pop("serialiser"))

        storage_specification = cls(
            engine=engine,
            chunk_size=chunk_size,
            serialiser=serialiser,
        )

        storage_specification.additional_properties = d
        return storage_specification

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.backtest_mode import BacktestMode
from ..models.classification_score_metric import ClassificationScoreMetric
from ..types import UNSET, Unset

T = TypeVar("T", bound="ClassificationBacktestConfig")


@attr.s(auto_attribs=True)
class ClassificationBacktestConfig:
    """ Properties for a stage  """

    backtest_mode: BacktestMode
    scale_factor: float
    score_metric_type: ClassificationScoreMetric
    start_timestamp_milliseconds: Union[Unset, None, int] = UNSET
    stop_timestamp_milliseconds: Union[Unset, None, int] = UNSET
    training_buffer_milliseconds: Union[Unset, None, int] = UNSET
    type: Union[Unset, str] = "classification_backtest"
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        backtest_mode = self.backtest_mode.value

        scale_factor = self.scale_factor
        score_metric_type = self.score_metric_type.value

        start_timestamp_milliseconds = self.start_timestamp_milliseconds
        stop_timestamp_milliseconds = self.stop_timestamp_milliseconds
        training_buffer_milliseconds = self.training_buffer_milliseconds
        type = self.type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "backtestMode": backtest_mode,
                "scaleFactor": scale_factor,
                "scoreMetricType": score_metric_type,
            }
        )
        if start_timestamp_milliseconds is not UNSET:
            field_dict["startTimestampMilliseconds"] = start_timestamp_milliseconds
        if stop_timestamp_milliseconds is not UNSET:
            field_dict["stopTimestampMilliseconds"] = stop_timestamp_milliseconds
        if training_buffer_milliseconds is not UNSET:
            field_dict["trainingBufferMilliseconds"] = training_buffer_milliseconds
        if type is not UNSET:
            field_dict["type"] = type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        backtest_mode = BacktestMode(d.pop("backtestMode"))

        scale_factor = d.pop("scaleFactor")

        score_metric_type = ClassificationScoreMetric(d.pop("scoreMetricType"))

        start_timestamp_milliseconds = d.pop("startTimestampMilliseconds", UNSET)

        stop_timestamp_milliseconds = d.pop("stopTimestampMilliseconds", UNSET)

        training_buffer_milliseconds = d.pop("trainingBufferMilliseconds", UNSET)

        type = d.pop("type", UNSET)

        classification_backtest_config = cls(
            backtest_mode=backtest_mode,
            scale_factor=scale_factor,
            score_metric_type=score_metric_type,
            start_timestamp_milliseconds=start_timestamp_milliseconds,
            stop_timestamp_milliseconds=stop_timestamp_milliseconds,
            training_buffer_milliseconds=training_buffer_milliseconds,
            type=type,
        )

        classification_backtest_config.additional_properties = d
        return classification_backtest_config

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import json
import time
from typing import List, Optional

import pandas as pd

from mindfoundry.client.horizon.generated import Client
from mindfoundry.client.horizon.generated.api.pipelines import (
    delete_pipeline_resource,
    get_feature_export_resource,
    get_feature_nodes_and_links_resource,
    get_pipeline_resource,
    post_classification_prediction_resource,
    post_on_demand_predict_resource,
    post_predict_resource,
    post_restart_pipeline_resource,
    post_run_pipeline_resource,
    put_all_pipeline_resource,
    put_pipeline_lock_resource,
    put_stage_creation_resource,
)
from mindfoundry.client.horizon.generated.models import (
    BlueprintType,
    BoolLocked,
    ClassificationPredictionForm,
    ClassificationPredictionResult,
    ExpertPredictForm,
    NodeAndLink,
    OnDemandPredictForm,
    Pipeline,
    PipelineSpecification,
    PredictResponse,
    RegressorType,
    StageSpecification,
    StageStatus,
)

from ..generated.types import File
from .utils import (
    DEFAULT_UPLOAD_OPTIONS,
    assert_success,
    data_frame_to_buffer,
    return_value_or_raise_error,
)


class PipelinesClient:
    def __init__(self, client: Client):
        self._client = client

    def get(self, pipeline_id: int) -> Pipeline:
        pipeline = get_pipeline_resource.sync_detailed(
            client=self._client,
            id=pipeline_id,
        )
        return return_value_or_raise_error(pipeline)

    def delete(self, pipeline_id: int) -> None:
        response = delete_pipeline_resource.sync_detailed(
            client=self._client,
            id=pipeline_id,
        )
        assert_success(response)

    def create_custom(self, dataset_id: int, name: str) -> Pipeline:
        pipeline = put_all_pipeline_resource.sync_detailed(
            client=self._client,
            json_body=PipelineSpecification(
                name=name,
                dataset_id=dataset_id,
                blueprint=BlueprintType.CUSTOM,
            ),
        )
        return return_value_or_raise_error(pipeline)

    def add_stage(self, pipeline_id: int, spec: StageSpecification) -> Pipeline:
        pipeline = put_stage_creation_resource.sync_detailed(
            client=self._client,
            id=pipeline_id,
            json_body=spec,
        )
        return return_value_or_raise_error(pipeline)

    def run(self, pipeline_id: int, wait_for_completion: bool = False) -> None:
        post_run_pipeline_resource.sync_detailed(client=self._client, id=pipeline_id)
        if wait_for_completion:
            self._wait_for_completion(pipeline_id)

    def restart(self, pipeline_id: int, wait_for_completion: bool = False) -> None:
        post_restart_pipeline_resource.sync_detailed(client=self._client, id=pipeline_id)
        if wait_for_completion:
            self._wait_for_completion(pipeline_id)

    def lock(self, pipeline_id: int) -> None:
        response = put_pipeline_lock_resource.sync_detailed(
            client=self._client,
            id=pipeline_id,
            json_body=BoolLocked(locked=True),
        )
        assert_success(response)

    def unlock(self, pipeline_id: int) -> None:
        response = put_pipeline_lock_resource.sync_detailed(
            client=self._client,
            id=pipeline_id,
            json_body=BoolLocked(locked=False),
        )
        assert_success(response)

    def predict_forecast(
        self,
        pipeline_id: int,
        horizon: int,
        regressor: RegressorType,
        new_data: Optional[pd.DataFrame] = None,
        filter_target_names: Optional[List[str]] = None,
    ) -> List[PredictResponse]:
        filter_target_names_as_json = (
            json.dumps(filter_target_names) if filter_target_names is not None else None
        )

        if new_data is None:
            prediction_response_list = post_on_demand_predict_resource.sync_detailed(
                client=self._client,
                id=pipeline_id,
                json_body=OnDemandPredictForm(
                    horizon=horizon,
                    regressor=regressor,
                    filter_target_names=filter_target_names_as_json,
                ),
            )

            return return_value_or_raise_error(prediction_response_list).predict_responses
        else:
            str_buffer = data_frame_to_buffer(new_data, "new_data.csv")
            prediction_response_list = post_predict_resource.sync_detailed(
                client=self._client,
                id=pipeline_id,
                multipart_data=ExpertPredictForm(
                    horizon=horizon,
                    regressor=regressor,
                    filter_target_names=filter_target_names_as_json,
                    options=DEFAULT_UPLOAD_OPTIONS,
                    file=File(
                        payload=str_buffer,
                        file_name=str_buffer.name,
                        mime_type="text/csv",
                    ),
                ),
            )

            return return_value_or_raise_error(prediction_response_list).predict_responses

    def predict_classification(
        self,
        pipeline_id: int,
        data: pd.DataFrame,
        target_label: str,
    ) -> ClassificationPredictionResult:
        """
        Given a new set of rows that match the training data, return a prediction for each row
        as well as feature importances.

        The model will be trained on the original data provided to the pipeline, and tested on
        the new data provided here.
        """
        str_buffer = data_frame_to_buffer(data, "data.csv")
        response = post_classification_prediction_resource.sync_detailed(
            client=self._client,
            id=pipeline_id,
            multipart_data=ClassificationPredictionForm(
                target_label=target_label,
                options=DEFAULT_UPLOAD_OPTIONS,
                file=File(
                    payload=str_buffer,
                    file_name=str_buffer.name,
                    mime_type="text/csv",
                ),
            ),
        )
        return return_value_or_raise_error(response)

    def download_features(
        self,
        pipeline_id: int,
        horizon: int,
        include_targets: bool,
    ) -> pd.DataFrame:
        pipeline = self.get(pipeline_id)
        last_stage = pipeline.stages[-1]

        features_response = get_feature_export_resource.sync_detailed(
            client=self._client,
            id=pipeline_id,
            stage_id=last_stage.id,
            horizon=horizon,
            include_targets=include_targets,
        )

        data_file = return_value_or_raise_error(features_response)
        return pd.read_csv(data_file.payload)

    def get_feature_graph(self, pipeline_id: int) -> List[NodeAndLink]:
        pipeline = self.get(pipeline_id)
        last_stage = pipeline.stages[-1]

        graph_response = get_feature_nodes_and_links_resource.sync_detailed(
            client=self._client,
            id=pipeline_id,
            stage_id=last_stage.id,
        )

        graph_data = return_value_or_raise_error(graph_response)
        return graph_data.nodes_and_links

    def _wait_for_completion(self, pipeline_id: int) -> None:
        pipeline = self.get(pipeline_id)

        while not all(stage.status == StageStatus.COMPLETE for stage in pipeline.stages):
            time.sleep(2)

            pipeline = self.get(pipeline_id)
            errored_stages = [
                stage for stage in pipeline.stages if stage.status == StageStatus.ERROR
            ]

            if len(errored_stages) > 0:
                raise RuntimeError(errored_stages[0].error_msg)



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.algo_execution_test_case import AlgoExecutionTestCase
from ..models.algo_test_spec_algoparams import AlgoTestSpecAlgoparams
from ..types import UNSET, Unset

T = TypeVar("T", bound="AlgoTestSpec")


@attr.s(auto_attribs=True)
class AlgoTestSpec:
    """  """

    algo_params: AlgoTestSpecAlgoparams
    test_case: Union[Unset, AlgoExecutionTestCase] = UNSET
    timeout_seconds: Union[Unset, int] = 30
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        algo_params = self.algo_params.to_dict()

        test_case: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.test_case, Unset):
            test_case = self.test_case.to_dict()

        timeout_seconds = self.timeout_seconds

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "algoParams": algo_params,
            }
        )
        if test_case is not UNSET:
            field_dict["testCase"] = test_case
        if timeout_seconds is not UNSET:
            field_dict["timeoutSeconds"] = timeout_seconds

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        algo_params = AlgoTestSpecAlgoparams.from_dict(d.pop("algoParams"))

        _test_case = d.pop("testCase", UNSET)
        test_case: Union[Unset, AlgoExecutionTestCase]
        if isinstance(_test_case, Unset):
            test_case = UNSET
        else:
            test_case = AlgoExecutionTestCase.from_dict(_test_case)

        timeout_seconds = d.pop("timeoutSeconds", UNSET)

        algo_test_spec = cls(
            algo_params=algo_params,
            test_case=test_case,
            timeout_seconds=timeout_seconds,
        )

        algo_test_spec.additional_properties = d
        return algo_test_spec

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class RebalanceWeeklyDay(str, Enum):
    NA = "na"
    MONDAY = "monday"
    TUESDAY = "tuesday"
    WEDNESDAY = "wednesday"
    THURSDAY = "thursday"
    FRIDAY = "friday"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.trading_algo_param_spec import TradingAlgoParamSpec
from ..types import UNSET, Unset

T = TypeVar("T", bound="TradingAlgo")


@attr.s(auto_attribs=True)
class TradingAlgo:
    """Holds code that will be executed each time the Trading Simulator requires an
    input.

    When a user creates an Algo, they define a list of parameters, with their
    names and types. These parameters will be set during the simulation, allowing Algos
    to be reused."""

    id: int
    name: str
    code: str
    author_user_id: int
    description: Union[Unset, str] = ""
    parameters_spec: Union[Unset, List[TradingAlgoParamSpec]] = UNSET
    is_builtin: Union[Unset, bool] = False
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        name = self.name
        code = self.code
        author_user_id = self.author_user_id
        description = self.description
        parameters_spec: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.parameters_spec, Unset):
            parameters_spec = []
            for parameters_spec_item_data in self.parameters_spec:
                parameters_spec_item = parameters_spec_item_data.to_dict()

                parameters_spec.append(parameters_spec_item)

        is_builtin = self.is_builtin

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "name": name,
                "code": code,
                "authorUserId": author_user_id,
            }
        )
        if description is not UNSET:
            field_dict["description"] = description
        if parameters_spec is not UNSET:
            field_dict["parametersSpec"] = parameters_spec
        if is_builtin is not UNSET:
            field_dict["isBuiltin"] = is_builtin

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        name = d.pop("name")

        code = d.pop("code")

        author_user_id = d.pop("authorUserId")

        description = d.pop("description", UNSET)

        parameters_spec = []
        _parameters_spec = d.pop("parametersSpec", UNSET)
        for parameters_spec_item_data in _parameters_spec or []:
            parameters_spec_item = TradingAlgoParamSpec.from_dict(parameters_spec_item_data)

            parameters_spec.append(parameters_spec_item)

        is_builtin = d.pop("isBuiltin", UNSET)

        trading_algo = cls(
            id=id,
            name=name,
            code=code,
            author_user_id=author_user_id,
            description=description,
            parameters_spec=parameters_spec,
            is_builtin=is_builtin,
        )

        trading_algo.additional_properties = d
        return trading_algo

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.lstm_backtest_stage_config_custommodelconfig import (
    LstmBacktestStageConfigCustommodelconfig,
)
from ..types import UNSET, Unset

T = TypeVar("T", bound="LstmBacktestStageConfig")


@attr.s(auto_attribs=True)
class LstmBacktestStageConfig:
    """ Properties for a stage  """

    epochs: int
    type: Union[Unset, str] = "lstm_backtest"
    dropout: Union[Unset, float] = UNSET
    number_of_units: Union[Unset, int] = UNSET
    time_steps: Union[Unset, int] = UNSET
    custom_model_config: Union[Unset, LstmBacktestStageConfigCustommodelconfig] = UNSET
    loss: Union[Unset, str] = UNSET
    optimizer: Union[Unset, str] = UNSET
    learning_rate: Union[Unset, float] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        epochs = self.epochs
        type = self.type
        dropout = self.dropout
        number_of_units = self.number_of_units
        time_steps = self.time_steps
        custom_model_config: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.custom_model_config, Unset):
            custom_model_config = self.custom_model_config.to_dict()

        loss = self.loss
        optimizer = self.optimizer
        learning_rate = self.learning_rate

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "epochs": epochs,
            }
        )
        if type is not UNSET:
            field_dict["type"] = type
        if dropout is not UNSET:
            field_dict["dropout"] = dropout
        if number_of_units is not UNSET:
            field_dict["numberOfUnits"] = number_of_units
        if time_steps is not UNSET:
            field_dict["timeSteps"] = time_steps
        if custom_model_config is not UNSET:
            field_dict["customModelConfig"] = custom_model_config
        if loss is not UNSET:
            field_dict["loss"] = loss
        if optimizer is not UNSET:
            field_dict["optimizer"] = optimizer
        if learning_rate is not UNSET:
            field_dict["learningRate"] = learning_rate

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        epochs = d.pop("epochs")

        type = d.pop("type", UNSET)

        dropout = d.pop("dropout", UNSET)

        number_of_units = d.pop("numberOfUnits", UNSET)

        time_steps = d.pop("timeSteps", UNSET)

        _custom_model_config = d.pop("customModelConfig", UNSET)
        custom_model_config: Union[Unset, LstmBacktestStageConfigCustommodelconfig]
        if isinstance(_custom_model_config, Unset):
            custom_model_config = UNSET
        else:
            custom_model_config = LstmBacktestStageConfigCustommodelconfig.from_dict(
                _custom_model_config
            )

        loss = d.pop("loss", UNSET)

        optimizer = d.pop("optimizer", UNSET)

        learning_rate = d.pop("learningRate", UNSET)

        lstm_backtest_stage_config = cls(
            epochs=epochs,
            type=type,
            dropout=dropout,
            number_of_units=number_of_units,
            time_steps=time_steps,
            custom_model_config=custom_model_config,
            loss=loss,
            optimizer=optimizer,
            learning_rate=learning_rate,
        )

        lstm_backtest_stage_config.additional_properties = d
        return lstm_backtest_stage_config

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import os

from setuptools import find_namespace_packages, setup

# Must be set by CI
version = os.environ["VERSION"]

# It's necessary to pin the httpx version, otherwise it will break file upload.
# Newer version throw an error when getting StringIO instead of a bytes-like object.
# See datasets_client.py:100 (_upload_csv function)
# See job: https://gitlab.mflocal.com/research/horizon/-/jobs/907935
install_requires = [
    "numpy",
    "pandas",
    "requests",
    "attrs",
    "httpx>=0.16.1,<0.19",
    "python-dateutil"
]

dev_extras = [
    "black==20.8b1",
    "click==8.0.3",
    "check-manifest==0.46",
    "coverage==5.5",
    "isort==5.8.0",
    "mypy==0.812",
    "pre-commit==2.11.1",
    "pylint==2.7.2",
    "pytest==6.2.4",
    "pytest-httpx==0.12.0",
    "pytest-timeout==1.4.2",
    "pytest-tldr==0.2.4",
    "pytest-asyncio==0.14.0",
    "bandit==1.7.0",
]

extras_require = {
    "dev": dev_extras,
    "all": install_requires + dev_extras,
}

with open("README.md", encoding="utf-8") as readme:
    long_description = readme.read()

setup(
    name="mindfoundry.client.horizon",
    version=version,
    license="MIT",
    description="Mind Foundry Horizon Client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://mindfoundry.ai/horizon",
    author="Mind Foundry Ltd",
    author_email="support@mindfoundry.ai",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_namespace_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=install_requires,
    extras_require=extras_require,
    namespace_packages=["mindfoundry"],
)



from enum import Enum


class BacktestMode(str, Enum):
    TRAINING_DATA = "training_data"
    VALIDATION_DATA = "validation_data"
    ALL_DATA = "all_data"
    CUSTOM_RANGE = "custom_range"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class CorrelationMethod(str, Enum):
    MUTUAL_INFO = "mutual_info"
    SPEARMAN = "spearman"
    PEARSON = "pearson"
    KENDALL = "kendall"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, List, Type, TypeVar, Union, cast

import attr

from ..models.rebalance_frequency import RebalanceFrequency
from ..models.rebalance_weekly_day import RebalanceWeeklyDay
from ..types import UNSET, Unset

T = TypeVar("T", bound="TradingSimulationOptions")


@attr.s(auto_attribs=True)
class TradingSimulationOptions:
    """  """

    targets: Union[Unset, List[str]] = UNSET
    start_timestamp: Union[Unset, None, int] = UNSET
    end_timestamp: Union[Unset, None, int] = UNSET
    starting_capital: Union[Unset, float] = 100000.0
    commission_fraction: Union[Unset, float] = 0.0
    allow_short_positions: Union[Unset, bool] = True
    max_gross_leverage: Union[Unset, float] = 1.0
    rebalance_frequency: Union[Unset, RebalanceFrequency] = RebalanceFrequency.DAILY
    rebalance_weekly_day: Union[Unset, RebalanceWeeklyDay] = RebalanceWeeklyDay.NA
    max_time_in_seconds: Union[Unset, int] = 86400
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        targets: Union[Unset, List[str]] = UNSET
        if not isinstance(self.targets, Unset):
            targets = self.targets

        start_timestamp = self.start_timestamp
        end_timestamp = self.end_timestamp
        starting_capital = self.starting_capital
        commission_fraction = self.commission_fraction
        allow_short_positions = self.allow_short_positions
        max_gross_leverage = self.max_gross_leverage
        rebalance_frequency: Union[Unset, str] = UNSET
        if not isinstance(self.rebalance_frequency, Unset):
            rebalance_frequency = self.rebalance_frequency.value

        rebalance_weekly_day: Union[Unset, str] = UNSET
        if not isinstance(self.rebalance_weekly_day, Unset):
            rebalance_weekly_day = self.rebalance_weekly_day.value

        max_time_in_seconds = self.max_time_in_seconds

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if targets is not UNSET:
            field_dict["targets"] = targets
        if start_timestamp is not UNSET:
            field_dict["startTimestamp"] = start_timestamp
        if end_timestamp is not UNSET:
            field_dict["endTimestamp"] = end_timestamp
        if starting_capital is not UNSET:
            field_dict["startingCapital"] = starting_capital
        if commission_fraction is not UNSET:
            field_dict["commissionFraction"] = commission_fraction
        if allow_short_positions is not UNSET:
            field_dict["allowShortPositions"] = allow_short_positions
        if max_gross_leverage is not UNSET:
            field_dict["maxGrossLeverage"] = max_gross_leverage
        if rebalance_frequency is not UNSET:
            field_dict["rebalanceFrequency"] = rebalance_frequency
        if rebalance_weekly_day is not UNSET:
            field_dict["rebalanceWeeklyDay"] = rebalance_weekly_day
        if max_time_in_seconds is not UNSET:
            field_dict["maxTimeInSeconds"] = max_time_in_seconds

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        targets = cast(List[str], d.pop("targets", UNSET))

        start_timestamp = d.pop("startTimestamp", UNSET)

        end_timestamp = d.pop("endTimestamp", UNSET)

        starting_capital = d.pop("startingCapital", UNSET)

        commission_fraction = d.pop("commissionFraction", UNSET)

        allow_short_positions = d.pop("allowShortPositions", UNSET)

        max_gross_leverage = d.pop("maxGrossLeverage", UNSET)

        _rebalance_frequency = d.pop("rebalanceFrequency", UNSET)
        rebalance_frequency: Union[Unset, RebalanceFrequency]
        if isinstance(_rebalance_frequency, Unset):
            rebalance_frequency = UNSET
        else:
            rebalance_frequency = RebalanceFrequency(_rebalance_frequency)

        _rebalance_weekly_day = d.pop("rebalanceWeeklyDay", UNSET)
        rebalance_weekly_day: Union[Unset, RebalanceWeeklyDay]
        if isinstance(_rebalance_weekly_day, Unset):
            rebalance_weekly_day = UNSET
        else:
            rebalance_weekly_day = RebalanceWeeklyDay(_rebalance_weekly_day)

        max_time_in_seconds = d.pop("maxTimeInSeconds", UNSET)

        trading_simulation_options = cls(
            targets=targets,
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            starting_capital=starting_capital,
            commission_fraction=commission_fraction,
            allow_short_positions=allow_short_positions,
            max_gross_leverage=max_gross_leverage,
            rebalance_frequency=rebalance_frequency,
            rebalance_weekly_day=rebalance_weekly_day,
            max_time_in_seconds=max_time_in_seconds,
        )

        trading_simulation_options.additional_properties = d
        return trading_simulation_options

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
